﻿//*****************************************************************************************
//** Name        : BiDiBLib                                                              **
//** Project     : DLL-Library for the handling of BiDiB-Messages via serial port        **
//** Version     : 0.0.1.1                                                               **
//** Autor       : Andreas Tillner 
//** Copyright   : 2013 (c) by Andreas Tillner
//** 
//** This source file is subject of the GNU general public license 3,
//** that is available at the world-wide-web at http://www.gnu.org/licenses/gpl.txt
//**
//**                   -------------------------------
//**                  // BBBB       DDDD       BBBB  \\
//**                 //  B   B  II  D   D  II  B   B  \\
//**                //   BBBB       D   D      BBBB    \\
//**                \\   B   B  II  D   D  II  B   B   //
//**                 \\  B   B  II  D   D  II  B   B  //
//**                  \\ BBBB   II  DDDD   II  BBBB  //
//**                   -------------------------------
//** This library based on the BiDiB-Standard by Wolfgang Kufer
//** Website: http://www.bidib.org
//** 
//** Last update : 14.11.2013                                                            
//** Changes     : 13.03.13 - added var MessageList, MsgFiFo, MsgTail, MsgHead           
//**                        - added public functions: 
//**                          Init ( int )
//**                          Init ()              
//**               22.03.13 - added Class BiDiBSerPort      
//**               23.03.13 - added var MsgMi, escape_hot, MsgFirst, BiDiBSerPort, MainForm, ProceedMSG                           
//**                        - added public functions: 
//**                          GetMessageNum(string BM) 
//**                          GetMsgString(byte BN)
//**                          SendMSGToSerial(byte[] msgb, string MB, bool log, bool sim)
//**                          SendMsgMultiParam(byte[] adr, string bm, int bnum, byte[] param, bool log, byte nu, bool sim)
//**                          SendMsgTwoParam(byte[] adr, string bm, byte param1, byte param2, bool log, byte nu, bool sim)
//**                          SendMsgOneParam(byte[] adr, string bm, byte param1, bool log, byte nu, bool sim)
//**                          SendMsgNoParam(byte[] adr, string bm, bool log, byte nu, bool sim)
//**                          HexByteToBox(byte[] data, int dl)
//**                          HexByteToBoxSub(byte[] data, int from, int dl)
//**               24.03.13 - added public functions:
//**                          InitSerialPortReg(string RegSubKey)
//**                          InitSerPort(string comport, int baud)
//**                          DataReceviedHandler(object sender, SerialDataReceivedEventArgs e)
//**                          ExtractMsg(byte[] msg)
//**                          Init(string MsgXML)
//**               26.03.13 - added var MsgCRCok[512] as bool
//**               12.05.13 - added public class for PlugIn-Interface
//**                          NodeFeatures, MsgEventArgs, TxtEventArgs
//**                        - added public functions:
//**                          SendByteMsg(byte[] param, bool log, byte nu, bool sim)
//**                          BuildMsgMultiParam(byte[] adr, string bm, int bnum, byte[] param, byte nu )
//**               24.07.13 - Update for new field in Message-XML <magic></magic>
//**               18.08.13 - Length of BiDiB-Message limited to 64 Byte
//**               11.09.13 - Change GetMessageNum(string BM) and GetMsgString(byte BN) to handle new BiDiBMessages.xml-Format
//**               25.10.13 - added var ProceedMSGRunning as bool - semafor true if ProceedMSG already running
//**               14.11.13 - string SendPlaneByteToSerial(byte[] msgb, string MB, bool log, bool sim)           
//**                                                                                     **
//*****************************************************************************************
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Linq;
using System.Text;
using System.IO.Ports;
using System.IO;
using System.Threading;
using System.Configuration;
using System.Timers;
using System.Reflection;
using System.Resources;
using System.Windows.Forms;

namespace BiDiBLib
{
    public class MsgEventArgs : EventArgs           // For PlugIn-Interface
    {
        public string _newmessagestring;
        public byte[] _newmessageparam;
        public byte[] _adr;
    }
    public class TxtEventArgs : EventArgs           // For PlugIn-Interface
    {
        public string _newmessage;
        public byte[] _adr;
        public bool _TimeStamp;
        public bool _LogToFileAndTextbox;
    }

    public class BiDiBSerialPort
    {           
        public SerialPort BiDiBPort = new SerialPort();
        public string ComPort;
        public int BaudRateIndex;
    }

    public class NodeFeature
    {
        public byte FetID;
        public byte value;
    }

    public class BiDiBMsgHandler
    {
        public DataSet MessageList = new DataSet(); //** Dataset with all messages from the Message-XML-File
        public byte[][] MsgFiFo = new byte[512][];
        public bool[] MsgCRCok = new bool[512];
        public byte[][] InputFiFo = new byte[512][];
        public int MsgTail;
        public int MsgHead; 
        public int InputTail;
        public int InputHead;
        public int MsgMaxLength;
        private string hexmsg;
        private int MsgCrc;
        private int MsgMi;
        private bool escape_hot;
        public bool ProceedMSGRunning;
        public bool BiDiBIsActive;
        public int MsgFirst;
        public BiDiBSerialPort BiDiBSerPort = new BiDiBSerialPort();
        public Form MainForm;
        public delegate void ProceedMSGDe();
        public ProceedMSGDe ProceedMSG;
        public delegate string SendMSGe(byte[] msgb, string MB, bool log, bool sim);
        public SendMSGe SendMSG;
        

        /**********************************************************************************
         ** Name        : InitSerialPortReg(string RegSubKey)
         ** Parameter   : string RegSubKey : Subkey in Registry under HKEY_LOCAL_USER
         ** Description : Read the config from Registry and set the parameter of the serial port.
         ** Return      : %%* = if the returnstring starts with %%, this is an error string
         **               All other strings are information or log.  
         **********************************************************************************/
        public string InitSerialPortReg(string RegSubKey)
        {
            Microsoft.Win32.RegistryKey SerKey;
            string rc;

            rc = "%%NO-InitSerialPortReg";

            SerKey = Microsoft.Win32.Registry.CurrentUser.OpenSubKey(RegSubKey, false);
            if (SerKey == null)
            {
                SerKey = Microsoft.Win32.Registry.CurrentUser.CreateSubKey(RegSubKey);
            }
            //********************************************************
            if ((SerKey.GetValue("Comport", null)) == null)
            {
                SerKey = Microsoft.Win32.Registry.CurrentUser.CreateSubKey(RegSubKey);
                SerKey.SetValue("Comport", "COM1");
            }
            //********************************************************
            if ((SerKey.GetValue("Baudrate", null)) == null)
            {
                SerKey = Microsoft.Win32.Registry.CurrentUser.CreateSubKey(RegSubKey);
                SerKey.SetValue("Baudrate", "2400");
            }
            //********************************************************
            string ComPort = Convert.ToString(SerKey.GetValue("Comport"));
            int baud = Convert.ToInt32(SerKey.GetValue("Baudrate"));
            SerKey.Close();
            rc = InitSerPort(ComPort, baud);
            return (rc);
        }

        /**********************************************************************************
        ** Name        : InitSerialPort(string comport, string baud)
        ** Parameter   : string comport : Comport as a string. COM5
        **             : in baud        : baudrate. 1000000, 115200 or 19200 alowed
        ** Description : Set the parameter of the serial port.
        ** Return      : string
        **               %%* = if the returnstring starts with %%, this is an error string
        **               All other strings are information or log.  
        **********************************************************************************/
        public string InitSerPort(string comport, int baud)
        {
            string rc;

            rc = "%%NO-InitSerPort";
            switch (baud)
            {
                case 19200:
                    BiDiBSerPort.BaudRateIndex = 2;
                    break;
                case 115200:
                    BiDiBSerPort.BaudRateIndex = 1;
                    break;
                case 1000000:
                    BiDiBSerPort.BaudRateIndex = 0;
                    break;
                default:
                    BiDiBSerPort.BaudRateIndex = 3;
                    break;
            }

            if (BiDiBSerPort.BiDiBPort.IsOpen == true)
            {
                BiDiBSerPort.BiDiBPort.Close();
                BiDiBSerPort.BiDiBPort.Dispose();
                rc = "%%INFO : " + BiDiBSerPort.BiDiBPort.PortName.ToString() + " was open and is closed now!!\n";
            }
            BiDiBSerPort.BiDiBPort.PortName = comport;
            BiDiBSerPort.BiDiBPort.BaudRate = baud;
            if (BiDiBSerPort.BaudRateIndex == 3)
            {
                rc = "%%ERROR: Please define COM-Port and Baudrate !!!\n";
            }
            else
            {
                BiDiBSerPort.BiDiBPort.PortName = comport;
                BiDiBSerPort.BiDiBPort.BaudRate = baud;
                BiDiBSerPort.BiDiBPort.ReadBufferSize = 8192;
                BiDiBSerPort.BiDiBPort.WriteBufferSize = 4096;
                BiDiBSerPort.BiDiBPort.Parity = Parity.None;
                BiDiBSerPort.BiDiBPort.StopBits = StopBits.One;
                BiDiBSerPort.BiDiBPort.DataBits = 8;
                BiDiBSerPort.BiDiBPort.Handshake = Handshake.RequestToSend;
                BiDiBSerPort.BiDiBPort.DataReceived += new SerialDataReceivedEventHandler(DataReceviedHandler);
                rc = "BiDiB at " + comport + " with " + baud + " Baud (" + BiDiBSerPort.BaudRateIndex + ")\n";
            }
            if ( ProceedMSG == null )
                rc = "%%ERROR: Please declare 'ProceedMSG' before open the COM-Port!!\n";
            return rc ;
        }
        /**********************************************************************************
         ** Name        : DataReceviedHandler(object sender, SerialDataReceivedEventArgs e)
         ** Parameter   : sender and SerialDataReceivedEventArgs e
         ** Description : Eventhandler for the serialport
         ** Return      : - 
         **********************************************************************************/
        private void DataReceviedHandler(object sender, SerialDataReceivedEventArgs e)
        {
            int bytes = BiDiBSerPort.BiDiBPort.BytesToRead;
            byte[] comBuffer = new byte[bytes];
            BiDiBSerPort.BiDiBPort.Read(comBuffer, 0, bytes);
            MainForm.BeginInvoke(new EventHandler(delegate { ExtractMsg(comBuffer); }));

        }

        /**********************************************************************************
         ** Name        : ExtractMsg(byte[] msg)
         ** Parameter   : byte[] msg   : array of bytes contains the data from serial port
         ** Description : build single messages from the datastream and save them to the Message FiFo
         ** Return      : -  
         **********************************************************************************/
        public void ExtractMsg(byte[] msg)
        {
            int i;
            int msglen;
            byte OneByte, escape_char, magic_char;


            i = 0;
            msglen = msg.Length;
            escape_char = 0xfd;
            magic_char = 0xfe;
            InputFiFo[InputHead] = new byte[msg.Length];
            if (BiDiBIsActive == true)
            {
                for (i = 0; i < msg.Length; i++)
                {
                    InputFiFo[InputHead][i] = msg[i];
                    OneByte = msg[i];
                    if (OneByte == magic_char)
                    {
                        escape_hot = false;
                        if (MsgCrc != 0)
                        {
                            MsgCrc = 0;
                            MsgCRCok[MsgHead] = false;      //*** CRC for Message in MsgFiFo[MsgHead] was not OK
                        }
                        else
                        {
                            MsgCRCok[MsgHead] = true;       //*** CRC for Message in MsgFiFo[MsgHead] was OK
                            MsgFiFo[MsgHead][0] = Convert.ToByte(MsgMi);
                        }
                        MsgMi = 1;
                        if (MsgFirst == 0)
                            MsgFirst = 1;
                        else
                        {
                            MsgHead++;
                            MsgCrc = 0;
                            if (MsgHead > 511)
                                MsgHead = 0;
                        }
                    }
                    else
                    {
                        if (OneByte == escape_char)
                            escape_hot = true;
                        else
                        {
                            if (escape_hot == true)
                            {
                                OneByte = Convert.ToByte(OneByte ^ 0x20);
                                escape_hot = false;
                            }
                            MsgCrc = crc_bists(Convert.ToByte(OneByte ^ MsgCrc));
                            MsgFiFo[MsgHead][MsgMi++] = OneByte;
                        }
                    }
                }
            }
            if (BiDiBIsActive == false)
            {
                int z = 0;
                for (i = 0; i < msg.Length; i++)
                {
                    if ((msg[i] == 0x0a) || (msg[i] == 0x0d))
                        InputFiFo[InputHead][z++] = msg[i]; 
                    if ((msg[i] >= 0x20) && (msg[i] <= 0x7e))
                        InputFiFo[InputHead][z++] = msg[i];
                }

            }
            if (msg.Length > 0)
                InputHead++;
            if (InputHead > 511)
                InputHead = 0;
            if ((MsgHead != MsgTail) || (InputHead != InputTail))
            {
                if ((ProceedMSG != null) && (ProceedMSGRunning == false))
                {
                    ProceedMSGRunning = true;   // Has to be set to false by ProceedMSG(), when it has finished
                    ProceedMSG();               //*** Must be defined in BiDiBMsgHandler.ProceedMSG
                }
            }
        }
        /**********************************************************************************
         ** Name        : SendMsgNoParam(byte[] adr, string bm, bool log, byte nu, bool sim)
         ** Parameter   : byte[] adr   : array of bytes contains the adress of the node
         **               string bm    : Message in plain text like MSG_BM_OCC
         **               bool log     : true = returns logoutput, false: returns status
         **               byte nu      : messagenumber
         **               bool sim     : true = simulation mode on, false = simulation mode off
         ** Description : build a message with one parameter and send it to the interface
         ** Return      : %%* = if the returnstring starts with %%, this is an error string
         **               All other strings are information or log.  
         **********************************************************************************/
        public string SendMsgNoParam(byte[] adr, string bm, bool log, byte nu, bool sim)
        {
            byte[] msgparam = new byte[2];
            string rc;

            rc = "%%NO";
            msgparam[0] = 0x00;
            rc = SendMsgMultiParam(adr, bm, 0, msgparam, log, nu, sim);
            return rc;
        }
        /**********************************************************************************
         ** Name        : SendMsgOneParam(byte[] adr, string bm, byte param1, bool log, byte nu, bool sim)
         ** Parameter   : byte[] adr   : array of bytes contains the adress of the node
         **               string bm    : Message in plain text like MSG_BM_OCC
         **               byte param1  : Byte of first parameter
         **               bool log     : true = returns logoutput, false: returns status
         **               byte nu      : messagenumber
         **               bool sim     : true = simulation mode on, false = simulation mode off
         ** Description : build a message with one parameter and send it to the interface
         ** Return      : %%* = if the returnstring starts with %%, this is an error string
         **               All other strings are information or log.  
         **********************************************************************************/
        public string SendMsgOneParam(byte[] adr, string bm, byte param1, bool log, byte nu, bool sim)
        {
            byte[] msgparam = new byte[2];
            string rc;

            rc = "%%NO";
            msgparam[0] = param1;
            rc = SendMsgMultiParam(adr, bm, 1, msgparam, log, nu, sim);
            return rc;
        }
        /**********************************************************************************
         ** Name        : SendMsgTwoParam(byte[] adr, string bm, byte param1, byte param2, bool log, byte nu, bool sim)
         ** Parameter   : byte[] adr   : array of bytes contains the adress of the node
         **               string bm    : Message in plain text like MSG_BM_OCC
         **               byte param1  : Byte of first parameter
         **               byte param2  : Byte of second parameter
         **               bool log     : true = returns logoutput, false: returns status
         **               byte nu      : messagenumber
         **               bool sim     : true = simulation mode on, false = simulation mode off
         ** Description : build a message with two parameters and send it to the interface
         ** Return      : %%* = if the returnstring starts with %%, this is an error string
         **               All other strings are information or log.  
         **********************************************************************************/
        public string SendMsgTwoParam(byte[] adr, string bm, byte param1, byte param2, bool log, byte nu, bool sim)
        {
            byte[] msgparam = new byte[2];
            string rc;

            rc = "%%NO";
            msgparam[0] = param1;
            msgparam[1] = param2;
            rc = SendMsgMultiParam(adr, bm, 2, msgparam, log, nu, sim);
            return rc;
        }
        /**********************************************************************************
         ** Name        : SendMsgMultiParam(byte[] adr, string bm, int bnum, byte[] param, bool log, byte nu, bool sim)
         ** Parameter   : byte[] adr   : array of bytes contains the adress of the node
         **               string bm    : Message in plain text like MSG_BM_OCC
         **               int bnum     : number of bytes of the parameter array
         **               byte[] param : array of byte with the paramters
         **               bool log     : true = returns logoutput, false: returns status
         **               byte nu      : messagenumber
         **               bool sim     : true = simulation mode on, false = simulation mode off
         ** Description : build a message with bnum parameters and send it to the interface
         ** Return      : %%* = if the returnstring starts with %%, this is an error string
         **               All other strings are information or log.  
         **********************************************************************************/
        public string SendMsgMultiParam(byte[] adr, string bm, int bnum, byte[] param, bool log, byte nu, bool sim)
        {
            byte[] msgb = new byte[64];
            int i, l, m, bcount;
            byte crc;
            string rc;

            rc = "-";
            i = 0;
            crc = 0;

            msgb[i++] = 0xfe;
            msgb[i++] = 0x00;
            msgb[i++] = adr[0];
            if (adr[0] > 0)
            {
                if (adr[1] > 0)
                {
                    msgb[i++] = adr[1];
                    if (adr[2] > 0)
                    {
                        msgb[i++] = adr[2];
                        if (adr[3] > 0)
                            msgb[i++] = adr[3];
                    }
                }
                msgb[i++] = 0;
            }
            msgb[i++] = nu;  //**** Messagenumber
            
            msgb[i++] = GetMessageNum(bm);
            if (msgb[i - 1] == 255)
                rc = "%%ERROR: Message '" + bm + "' not found in BIDIB-Messages.xml\n";
            else
            {
                bcount = 0;
                while ((bcount < bnum) && (i < 62))   // 0-62 + crc = 64 Byte
                    msgb[i++] = param[bcount++];

                msgb[1] = Convert.ToByte(i);
                msgb[1] -= 2;
                m = msgb[1] + 1;
                for (l = 1; l <= m; l++)
                    crc = crc_bists(Convert.ToByte(msgb[l] ^ crc));
                msgb[i++] = crc;
                msgb[i++] = 0xfe;

                if ( SendMSG != null )
                  rc = SendMSG(msgb, bm, log, sim);
                else
                  rc = SendMSGToSerial(msgb, bm, log, sim);
            }
            return rc;
        }
        /**********************************************************************************
         ** Name        : SendByteMsg(byte[] param, bool log, byte nu, bool sim)
         ** Parameter   : byte[] param : array of byte with the Message
         **                              0 = number of bytes, incl. 0 and 1
         **                              1 = position of msg-number
         **                              2 - n = Messagebytes
         **               bool log     : true = returns logoutput, false: returns status
         **               byte nu      : messagenumber
         **               bool sim     : true = simulation mode on, false = simulation mode off
         ** Description : build a message with bnum parameters and send it to the interface
         ** Return      : %%* = if the returnstring starts with %%, this is an error string
         **               All other strings are information or log.  
         **********************************************************************************/
        public string SendByteMsg(byte[] msg, bool log, byte nu, bool sim)
        {
            byte[] msgb = new byte[128];
            int i, l, m;
            byte crc;
            string rc;

            rc = "-";
            i = 0;
            crc = 0;
            m = msg[0]-2;

            msg[msg[1]] = nu;
            for (i = 0; i <= m; i++)
            {
                msgb[i] = msg[i+3];
            }
            crc = 0;
            m = msgb[1] + 1;
            for (l = 1; l <= m; l++)
                crc = crc_bists(Convert.ToByte(msgb[l] ^ crc));
            msgb[m+1] = crc;

            string bm = GetMsgString(msg[msg[2]]);
            if (bm.Substring(0,2).CompareTo("%%") == 0)
                rc = "%%ERROR: Message-Nr. '" + String.Format("{0:X2}",msg[msg[2]]) + "' not found in BIDIB-Messages.xml\n";
            else
            {
                if (SendMSG != null)
                    rc = SendMSG(msgb, bm, log, sim);
                else
                    rc = SendMSGToSerial(msgb, bm, log, sim);
            }
            return rc;
        }
        /**********************************************************************************
         ** Name        : BuildMsgMultiParam(byte[] adr, string bm, int bnum, byte[] param, byte nu )
         ** Parameter   : byte[] adr   : array of bytes contains the adress of the node
         **               string bm    : Message in plain text like MSG_BM_OCC
         **               int bnum     : number of bytes of the parameter array
         **               byte[] param : array of byte with the paramters
         **               byte nu      : messagenumber
         ** Description : build a message with bnum parameters and return the byte-array with the message
         ** Return      : byte[] msgb : 0 = number of bytes, incl. 0,1 and 2
         **                             1 = position of msg-counter
         **                             2 = position of msg-number
         **                             3 - n = Messagebytes
         **               The build message is only valid, if msgb[0] > 0
         **********************************************************************************/
        public byte[] BuildMsgMultiParam(byte[] adr, string bm, int bnum, byte[] param, byte nu)
        {
            byte[] msgb = new byte[128];
            int i,  m, bcount;
            byte crc;

            i = 3;
            crc = 0;

            msgb[i++] = 0xfe;
            msgb[i++] = 0x00;
            msgb[i++] = adr[0];
            if (adr[0] > 0)
            {
                if (adr[1] > 0)
                {
                    msgb[i++] = adr[1];
                    if (adr[2] > 0)
                    {
                        msgb[i++] = adr[2];
                        if (adr[3] > 0)
                            msgb[i++] = adr[3];
                    }
                }
                msgb[i++] = 0;
            }
            msgb[1] = Convert.ToByte(i);
            msgb[i++] = 0x00;  //**** Messagecounter

            msgb[2] = Convert.ToByte(i);
            msgb[i++] = GetMessageNum(bm);
            if (msgb[i - 1] == 255)
            {
                msgb[0] = 0x00;
                msgb[1] = 0x01; //*** Message not found
                msgb[2] = 0x00;
            }
            else
            {

                if (msgb[i - 1] == 255)
                {
                    msgb[0] = 0x00;
                }
                else
                {
                    for (bcount = 0; bcount < bnum; bcount++)
                        msgb[i++] = param[bcount];
                    msgb[4] = Convert.ToByte(i - 5);
                    m = msgb[4] + 5;
                    crc = 0x00;

                    msgb[i++] = crc;
                    msgb[i++] = 0xfe;

                }
                msgb[0] = Convert.ToByte(i);
            }
            return msgb;
        }
        /**********************************************************************************
         ** Name        : GetMessageNum(string bm)
         ** Parameter   : string bm
         **               bm = Message in plain text like MSG_BM_OCC
         ** Description : Find a Message in DataSet and return the Msg.-Number as Byte
         ** Return      : 255 = Message not found
         **               0 - 254 = messagenumber as a byte            
         **********************************************************************************/
        public byte GetMessageNum(string BM)
        {
            DataRow[] foundrow;
            string SelStr;
            byte MessageNum = 255; ;

            SelStr = "Message = '" + BM + "'";
            foundrow = MessageList.Tables["Msg"].Select(SelStr);
            if (foundrow.Length > 0)
                MessageNum = byte.Parse(Convert.ToString(foundrow[0][2]), System.Globalization.NumberStyles.HexNumber, System.Globalization.CultureInfo.InvariantCulture);
            else
                MessageNum = 255;
            return MessageNum;
        }
        /**********************************************************************************
         ** Name        : GetMsgString(byte BN)
         ** Parameter   : string bm
         **               bm : Messagenumber in byte
         ** Description : Find a Msg.-Bumber in DataSet and return the Message as String
         ** Return      : 255 = Message not found
         **               0 - 254 = messagenumber as a byte            
         **********************************************************************************/
        public string GetMsgString(byte BN)
        {
            DataRow[] foundrow;
            string SelStr;
            string MsgStr = "-vvv-";

            SelStr = "Messagenum = '" + String.Format("{0:X2}", BN) + "'";
            foundrow = MessageList.Tables["Msg"].Select(SelStr);
            if (foundrow.Length > 0)
                MsgStr = Convert.ToString(foundrow[0][3]);
            else
                MsgStr = "%%ERROR: "+SelStr + "\n";

            return MsgStr;
        }
        /**********************************************************************************
         ** Name        : SendMSGToSerial(byte[] msgb, string MB, bool log, bool sim)
         ** Parameter   : byte[] msgb : array of bytes contains the message
         **               string bm   : Message in plain text like MSG_BM_OCC
         **               bool log    : true: returns logoutput, false: returns status
         **               bool sim    : true = simulation mode on, false = simulation mode off
         ** Description : Find a Message in DataSet and return the Msg.-Number as Byte
         ** Return      : -       = Message not send
         **               OK      = Message was send            
         **               logtext = Text for Logfile
         **********************************************************************************/
        public string SendMSGToSerial(byte[] msgb, string MB, bool log, bool sim)
        {
            byte[] msgbn = new byte[128];
            int msglen, i, j, inew;
            string rc;

            i = 0;
            inew = 0;
            rc = "-";

            msglen = msgb[1];
            msgbn[inew++] = msgb[i++];
            for (j = 0; j < msglen + 2; j++)
            {
                if ((msgb[i] == 0xFE) || (msgb[i] == 0xFD))
                {
                    msgbn[inew++] = 0xFD;
                    msgbn[inew++] = Convert.ToByte(msgb[i++] ^ 0x20);
                }
                else
                    msgbn[inew++] = msgb[i++];
            }
            msgbn[inew++] = msgb[i++];
            if ((BiDiBSerPort.BiDiBPort.IsOpen) || (sim == true))
            {
                if (sim == false)
                    BiDiBSerPort.BiDiBPort.Write(msgbn, 0, inew);
                if (log == true)
                    rc = "OUT -- : " + String.Format("{0,-25}", MB) + " " + HexByteToBox(msgbn, inew) + "\n";
                else
                    rc = "OK";
            }
            else
                rc = "%%ERROR: Serial Port NOT connected or open\n";

            return rc;
        }
        /**********************************************************************************
         ** Name        : SendPlaneByteToSerial(byte[] msgb, string MB, bool log, bool sim)
         ** Parameter   : byte[] msgb : array of bytes contains the message
         **               string bm   : Message in plain text like "Suppentopf"
         **               bool log    : true: returns logoutput, false: returns status
         **               bool sim    : true = simulation mode on, false = simulation mode off
         ** Description : Find a Message in DataSet and return the Msg.-Number as Byte
         ** Return      : -       = Message not send
         **               OK      = Message was send            
         **               logtext = Text for Logfile
         **********************************************************************************/
        public string SendPlaneByteToSerial(byte[] msgb, string MB, bool log, bool sim)
        {
            string rc;
            rc = "-";
            if ((BiDiBSerPort.BiDiBPort.IsOpen) || (sim == true))
            {
                if (sim == false)
                    BiDiBSerPort.BiDiBPort.Write(msgb, 0, msgb.Length);
                if (log == true)
                    rc = "ASCII -: " + String.Format("{0,-25}", MB) + " " + HexByteToBox(msgb, msgb.Length) + " \n";
                else
                    rc = "OK";
            }
            else
                rc = "%%ERROR: Serial Port NOT connected or open\n";

            return rc;
        }
        /**********************************************************************************
         ** Name        : Init ( int msgsize, string MsgXML )
         ** Parameter   : int msgsize   = max. size of a single message in byte
         **               string MsgXML = Full Path to the XML-File which contains the BiDiB-Messages
         ** Description : - Reading the Messagefile BiDiB-Messages.xml
         **               - Init the Message FiFo
         ** Return      : 1 >> OK
         **               2 >> No Message found in BiDiB-Messages.xml
         **********************************************************************************/
        public byte Init(int msgsize, string MsgXML)
        {
            byte rc = 0x01;                             //*** Default retruncode 0x01 = OK
            MsgMaxLength = msgsize;
            //***********************************************************//
            //****** Load the Message-XML-File into Dataset *************//
            MessageList.ReadXml(MsgXML);
            if (MessageList.Tables.Count < 1)           //*** Dataset still empty?
                rc = 0x02;                              //*** Error 0x02 = No Message in Dataset
            //***********************************************************//
            //****** Declare MessageFIFO ********************************//
            int f;
            for (f = 0; f < 512; f++)
                MsgFiFo[f] = new byte[msgsize];
            MsgTail = 0;
            MsgHead = 0;
            //***********************************************************//
            //****** Declare InputFIFO **********************************//
            InputTail = 0;
            InputHead = 0;
            //***********************************************************//
            escape_hot = false;
            MsgCrc = 0;
            MsgFirst = 0;
            MsgMi = 0;
            hexmsg = "-: ";
            ProceedMSG = null;
            SendMSG = null;

            return rc;
        }

        /**********************************************************************************
         ** Name        : Init ( void )
         ** Parameter   : -
         ** Description : - Reading the Messagefile BiDiB-Messages.xml
         **               - Init the Message FiFo
         **               - Call Init(int) with the default size of 128 Byte 
         **                 and the default Message-XML-File
         **********************************************************************************/
        public byte Init()
        {
            byte rc = 0x01;
            rc = Init(128, "data\\BiDiB-Messages.xml");
            return rc;
        }
        /**********************************************************************************
         ** Name        : Init ( string MsgXML )
         ** Parameter   : string MsgXML = Full Path to the XML-File which contains the BiDiB-Messages
         ** Description : - Reading the Messagefile from MsgXML
         **               - Init the Message FiFo
         **               - Init the Input FiFo
         **               - Call Init(int) with the default size of 128 Byte
         **********************************************************************************/
        public byte Init(string MsgXML)
        {
            byte rc = 0x01;
            rc = Init(128, MsgXML);
            return rc;
        }
        /**********************************************************************************
         ** Name        : crc_bists ( byte )
         ** Parameter   : byte ch
         **               ch = byte which crc will return 
         ** Description : get the crc from byte array, ch is the index.
         ** Return      : byte as crc from ch
         **               
         **********************************************************************************/
        public byte crc_bists(byte ch)
        {
            byte[] crc = new byte[] { 0x00, 0x5e, 0xbc, 0xe2, 0x61, 0x3f, 0xdd, 0x83,
                                      0xc2, 0x9c, 0x7e, 0x20, 0xa3, 0xfd, 0x1f, 0x41,
                                      0x9d, 0xc3, 0x21, 0x7f, 0xfc, 0xa2, 0x40, 0x1e,
                                      0x5f, 0x01, 0xe3, 0xbd, 0x3e, 0x60, 0x82, 0xdc,
                                      0x23, 0x7d, 0x9f, 0xc1, 0x42, 0x1c, 0xfe, 0xa0,
                                      0xe1, 0xbf, 0x5d, 0x03, 0x80, 0xde, 0x3c, 0x62,
                                      0xbe, 0xe0, 0x02, 0x5c, 0xdf, 0x81, 0x63, 0x3d,
                                      0x7c, 0x22, 0xc0, 0x9e, 0x1d, 0x43, 0xa1, 0xff,
                                      0x46, 0x18, 0xfa, 0xa4, 0x27, 0x79, 0x9b, 0xc5,
                                      0x84, 0xda, 0x38, 0x66, 0xe5, 0xbb, 0x59, 0x07,
                                      0xdb, 0x85, 0x67, 0x39, 0xba, 0xe4, 0x06, 0x58,
                                      0x19, 0x47, 0xa5, 0xfb, 0x78, 0x26, 0xc4, 0x9a,
                                      0x65, 0x3b, 0xd9, 0x87, 0x04, 0x5a, 0xb8, 0xe6,
                                      0xa7, 0xf9, 0x1b, 0x45, 0xc6, 0x98, 0x7a, 0x24,
                                      0xf8, 0xa6, 0x44, 0x1a, 0x99, 0xc7, 0x25, 0x7b,
                                      0x3a, 0x64, 0x86, 0xd8, 0x5b, 0x05, 0xe7, 0xb9,
                                      0x8c, 0xd2, 0x30, 0x6e, 0xed, 0xb3, 0x51, 0x0f,
                                      0x4e, 0x10, 0xf2, 0xac, 0x2f, 0x71, 0x93, 0xcd,
                                      0x11, 0x4f, 0xad, 0xf3, 0x70, 0x2e, 0xcc, 0x92,
                                      0xd3, 0x8d, 0x6f, 0x31, 0xb2, 0xec, 0x0e, 0x50,
                                      0xaf, 0xf1, 0x13, 0x4d, 0xce, 0x90, 0x72, 0x2c,
                                      0x6d, 0x33, 0xd1, 0x8f, 0x0c, 0x52, 0xb0, 0xee,
                                      0x32, 0x6c, 0x8e, 0xd0, 0x53, 0x0d, 0xef, 0xb1,
                                      0xf0, 0xae, 0x4c, 0x12, 0x91, 0xcf, 0x2d, 0x73,
                                      0xca, 0x94, 0x76, 0x28, 0xab, 0xf5, 0x17, 0x49,
                                      0x08, 0x56, 0xb4, 0xea, 0x69, 0x37, 0xd5, 0x8b,
                                      0x57, 0x09, 0xeb, 0xb5, 0x36, 0x68, 0x8a, 0xd4,
                                      0x95, 0xcb, 0x29, 0x77, 0xf4, 0xaa, 0x48, 0x16,
                                      0xe9, 0xb7, 0x55, 0x0b, 0x88, 0xd6, 0x34, 0x6a,
                                      0x2b, 0x75, 0x97, 0xc9, 0x4a, 0x14, 0xf6, 0xa8,
                                      0x74, 0x2a, 0xc8, 0x96, 0x15, 0x4b, 0xa9, 0xf7,
                                      0xb6, 0xe8, 0x0a, 0x54, 0xd7, 0x89, 0x6b, 0x35,};

            return (crc[ch]);
        }
        /******************************************************************/
        /******************************************************************/
        /******************************************************************/

        public string HexByteToBoxSub(byte[] data, int from, int dl)
        {
            string sdata = "";
            string tdata = "";
            int i, len;

            len = data.Length;
            len = from + dl;
            for (i = from; i <= len; i++)
            {
                tdata = String.Format("{0:X2}", data[i]); //System.Convert.ToString(data[i], 16));
                sdata = sdata + tdata + " ";
            }
            return sdata;
        }
        /******************************************************************/
        /******************************************************************/
        /******************************************************************/

        public string HexByteToBox(byte[] data, int dl)
        {
            string sdata = "";
            string tdata = "";
            int i, len;

            len = data.Length;
            if (dl > 0)
                len = dl;
            for (i = 0; i < len; i++)
            {
                tdata = String.Format("{0:X2}", data[i]); //System.Convert.ToString(data[i], 16));
                sdata = sdata + tdata + " ";
            }
            return sdata;
        }

        public void HexToArray(string hexString, byte[] bdata)
        {
/*            if (hexString.Length % 2 != 0)
            {
                throw new ArgumentException(String.Format("The binary key cannot have an odd number of digits: {0}", hexString));
            }
            */
            int il = hexString.Length/2;
            
            for (int index = 0; index < il; index++)
            {
                string byteValue = hexString.Substring(index*2,2);
                bdata[index] = TwoCharStringToByte(byteValue);
            }

        }

        public byte TwoCharStringToByte(string constr)
        {
            int b1, b2;
            b1 = 0;
            b2 = 0;
            switch (constr[0])
            {
                case '0': b1 = 0;
                    break;
                case '1': b1 = 1;
                    break;
                case '2': b1 = 2;
                    break;
                case '3': b1 = 3;
                    break;
                case '4': b1 = 4;
                    break;
                case '5': b1 = 5;
                    break;
                case '6': b1 = 6;
                    break;
                case '7': b1 = 7;
                    break;
                case '8': b1 = 8;
                    break;
                case '9': b1 = 9;
                    break;
                case 'a': b1 = 10;
                    break;
                case 'b': b1 = 11;
                    break;
                case 'c': b1 = 12;
                    break;
                case 'd': b1 = 13;
                    break;
                case 'e': b1 = 14;
                    break;
                case 'f': b1 = 15;
                    break;
                case 'A': b1 = 10;
                    break;
                case 'B': b1 = 11;
                    break;
                case 'C': b1 = 12;
                    break;
                case 'D': b1 = 13;
                    break;
                case 'E': b1 = 14;
                    break;
                case 'F': b1 = 15;
                    break;
                default: b1 = 0x0f;
                    break;

            }

            switch (constr[1])
            {
                case '0': b2 = 0;
                    break;
                case '1': b2 = 1;
                    break;
                case '2': b2 = 2;
                    break;
                case '3': b2 = 3;
                    break;
                case '4': b2 = 4;
                    break;
                case '5': b2 = 5;
                    break;
                case '6': b2 = 6;
                    break;
                case '7': b2 = 7;
                    break;
                case '8': b2 = 8;
                    break;
                case '9': b2 = 9;
                    break;
                case 'a': b2 = 10;
                    break;
                case 'b': b2 = 11;
                    break;
                case 'c': b2 = 12;
                    break;
                case 'd': b2 = 13;
                    break;
                case 'e': b2 = 14;
                    break;
                case 'f': b2 = 15;
                    break;
                case 'A': b2 = 10;
                    break;
                case 'B': b2 = 11;
                    break;
                case 'C': b2 = 12;
                    break;
                case 'D': b2 = 13;
                    break;
                case 'E': b2 = 14;
                    break;
                case 'F': b2 = 15;
                    break;
                default: b2 = 0x0f;
                    break;
            }
            b1 = b1 *16 + b2;
            return Convert.ToByte(b1);
        }
    }
}

